"""WorkToy - SYM - SymSpace
Special namespace class used by the SYM module."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.core import Bases
from worktoy.metaclass import AbstractNameSpace
from worktoy.sym import SYM


class SymSpace(AbstractNameSpace):
  """WorkToy - SYM - SymSpace
  Special namespace class used by the SYM module."""

  def __init__(self, name: str, bases: Bases, **kwargs) -> None:
    AbstractNameSpace.__init__(self, name, bases, **kwargs)
    self._name = name
    self._bases = bases
    self._symbolicBaseclass = None
    self._symBase = None
    self._nameSpace = None
    self._instanceSpace = None
    self._keywordSym = kwargs.get('symNames', [])

  def getModuleName(self) -> str:
    """Getter-function for the module name"""
    return 'WorkToy - SYM'

  def _getKeySyms(self) -> dict:
    return {k: SYM.auto() for k in self._keywordSym}

  def _setSymbolicBaseclass(self, cls: type) -> None:
    if self._symbolicBaseclass is not None:
      from worktoy.waitaminute import UnavailableNameException
      raise UnavailableNameException(
        '_symbolicBaseclass', self._symbolicBaseclass, cls)
    for base in self._bases:
      if getattr(base, '__symbolic_baseclass__', False):
        self._symbolicBaseClass = base
        return

  def _getSym(self) -> type:
    """Getter-function for the symbolic baseclass."""
    if self._symbolicBaseclass is None:
      from worktoy.waitaminute import MissingArgumentException
      raise MissingArgumentException('_symbolicBaseclass')
    return self._symbolicBaseclass

  def _validateInstance(self, _: str, val: Any) -> bool:
    """This method is responsible for deciding if an entry in the
    namespace indicates a symbolic instance. A subclass may reimplement
    this method to customize instance selection. The default behaviour is
    to look for the '__symbolic_instance__' flag on the instance. It does
    not use the key."""
    return True if getattr(val, '__symbolic_instance__', False) else False

  def _parseData(self, ) -> None:
    """Parses the data removing the instance creations in the class body."""
    self._instanceSpace = {}
    self._nameSpace = {}
    contentItems = [pair for pair in dict.items(self)]
    keyWordItems = [pair for pair in self._getKeySyms().items()]
    for (key, val) in [*contentItems, *keyWordItems]:
      key = key.replace(' ', '_').replace('.', '').lower()
      if self._validateInstance(key, val):
        self._instanceSpace |= {key: val}
      self._nameSpace |= {key: val}

  def parseInstance(self, *args, **kwargs) -> Any:
    """Parses the arguments to a particular instance. """
    raise NotImplementedError

  def getNameSpace(self, **kwargs) -> dict:
    """Creates a special dictionary with instance creations found in the
    class body removed. During the __new__ method, metaclasses should
    transmit this special dictionary on to the super call. """
    if self._nameSpace is None:
      if kwargs.get('_recursion', False):
        creator = self._parseData
        varType = dict
        varName = '_nameSpace'
        from worktoy.waitaminute import RecursiveCreateGetError
        raise RecursiveCreateGetError(creator, varType, varName)
      self._parseData()
      return self.getNameSpace(_recursion=True)
    return self._nameSpace

  def getInstanceSpace(self, **kwargs) -> dict:
    """Getter-function for the dictionary of the named symbolic instances."""
    if self._instanceSpace is None:
      if kwargs.get('_recursion', False):
        creator = self._instanceSpace
        varType = dict
        varName = '_instances'
        from worktoy.waitaminute import RecursiveCreateGetError
        raise RecursiveCreateGetError(creator, varType, varName)
      self._parseData()
      return self.getInstanceSpace(_recursion=True)
    return self._instanceSpace



"""WorkToy - WaitAMinute
Decorator-based custom exception"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from ._except_space import ExceptSpace
from ._meta_xcept import MetaXcept
from ._missing_argument_exception import MissingArgumentException
from ._type_support_error import TypeSupportError
from ._meta_type_support_error import MetaTypeSupportError
from ._unavailable_name import UnavailableNameException
from ._unsupported_subclass_exception import UnsupportedSubclassException
from ._recursive_create_get_error import RecursiveCreateGetError



"""WorkToy - Core - CoreClass
Provides utilities accessible by subclassing this CoreClass or its
subclasses.
Typically, subclasses would inherit from DefaultClass and add mixin
classes as appropriate."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Optional


class CoreClass:
  """WorkToy - Core - CoreClass
  Provides utilities accessible by subclassing this CoreClass or its
  subclasses."""

  def __init__(self, *args, **kwargs) -> None:
    self._args = args
    self._kwargs = kwargs

  def maybe(self, *args, ) -> Any:
    """
  The maybe function returns the first argument that is not None.

  :param *args: Pass in a tuple of arguments to the function
  :return: The first argument that is not 'None'
  """
    for arg in args:
      if arg is not None:
        return arg

  def maybeType(self, cls: type, *args) -> Any:
    """
    Returns the first positional argument belonging to the type given in
    the 'cls' argument.

    :param cls: type: Specify that the first parameter is a type
    :param *args: Pass a variable number of arguments to the function
    :return: The first argument that is an instance of 'cls'
    """
    for arg in args:
      if isinstance(arg, cls):
        return arg

  def maybeTypes(self, cls: type, *args, **kwargs) -> list:
    """
    Returns each positional argument belonging to the type given in the
    'cls' argument. Use keyword arguments 'pad' and 'padChar' to set a
    padding and a padding character. For example:
      maybeTypes(int, 'a', 1, 1.5, 2, pad=4, padChar=-1)
      >>> [1, 2, -1, -1]
    :param cls: type: Specify the type of class that we are looking for
    :param *args: Pass in a list of arguments
    :param **kwargs: Pass keyword arguments to the function
    :return: A list of all the arguments that are instances
    """
    out = [arg for arg in args if isinstance(arg, cls)]
    while len(out) < kwargs.get('pad', 0):
      out.append(kwargs.get('padChar', None))
    return out

  def searchKey(self, *keys, **kwargs) -> Any:
    """
    This method searches the positional arguments for an argument at each
    given key.
    :param *keys: Pass in a list of keys to search for
    :param **kwargs: Pass in keyword arguments to the function
    :return: The first value that is not none
    """
    for key in keys:
      val = kwargs.get(key, None)
      if val is not None:
        return val

  def searchKeys(self, *keys, **kwargs) -> Optional[list]:
    """Returns a list of 'kwargs' matching a key."""
    out = []
    if not keys:
      return None
    for key in keys:
      val = kwargs.get(key, None)
      if val is not None:
        out.append(val)
    return out or None

  def _maybeKey(self, allInstances: bool, *args, **kwargs) -> Any:
    """Finds the first object in kwargs that matches a given key. Provide
    keys as positional arguments of stringtype. Optionally provide a
    'type' in the positional arguments to return only an object of that
    type."""
    out = []
    if not kwargs:
      return out
    types = [arg for arg in args if isinstance(arg, type)]
    keys = [arg for arg in args if isinstance(arg, str)]
    keyVals = self.searchKeys(*keys, **kwargs)
    if keyVals is None:
      return keyVals
    if not types:
      return keyVals if allInstances else keyVals[0]
    out = []
    for cls in types:
      for val in keyVals:
        if isinstance(val, cls):
          if not allInstances:
            return val
          out.append(val)
    if allInstances:
      return out

  def maybeKey(self, *args, **kwargs) -> Any:
    """Finds the first object in kwargs that matches a given key and
    belongs to a given type."""
    return self._maybeKey(False, *args, **kwargs) or None

  def maybeKeys(self, *args, **kwargs) -> Any:
    """Finds all objects in kwargs matching any of the given keys and
    belonging to any given types. If no types are given, no type check is
    performed."""
    return self._maybeKey(True, *args, **kwargs) or None

  def empty(self, *args) -> bool:
    """Returns True if every positional argument is None. The method
    returns True when receiving no positional arguments. Otherwise, the
    method returns True."""
    if not args:
      return True
    for arg in args:
      if arg is not None:
        return False
    return True

  def plenty(self, *args) -> bool:
    """The method returns True if no positional argument is None.
    Otherwise, the method returns False. """
    if not args:
      return True
    for arg in args:
      if arg is None:
        return False
    return True

  def getArgs(self) -> list:
    """Getter-function for the list of positional arguments."""
    if self._args is None:
      return []
    return [arg for arg in self._args]

  def getKwargs(self) -> dict:
    """Getter-function for the dictionary of keyword arguments."""
    return self._kwargs

  def setArgs(self, *args) -> None:
    """Setter-function for the positional arguments"""
    self._args = args

  def setKwargs(self, **kwargs) -> None:
    """Setter function for the keyword arguments"""
    self._kwargs = kwargs

  def setAllArgs(self, *args, **kwargs) -> None:
    """Setting both positional arguments and keyword arguments."""
    self._args, self._kwargs = self.setArgs(*args), self.setKwargs(**kwargs)



"""WorkToy - Wait A Minute! - MetaTypeSupportError
Indicates that an instance is not a member of class derived from the
correct metaclass."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.waitaminute import MetaXcept


class MetaTypeSupportError(MetaXcept):
  """WorkToy - Wait A Minute! - MetaTypeSupportError
  Indicates that an instance is not a member of class derived from the
  correct metaclass."""

  def __init__(self, expMetaClass: type, actualValue: Any, argName: str,
               *args, **kwargs) -> None:
    MetaXcept.__init__(self, expMetaClass, actualValue, argName)
    self._expMetaClass = self.pretty(expMetaClass)
    self._actualValue = self.pretty(actualValue)
    self._actualType = self.pretty(actualValue.__class__)
    self._actualMetaClass = self.pretty(self._actualType.__class__)
    self._argName = self.pretty(argName)

  def __str__(self, ) -> str:
    header = MetaXcept.__str__(self)
    expMcls = self._expMetaClass
    actVal = self._actualValue
    actCls = self._actualType
    actMcls = self._actualMetaClass
    argName = self._argName
    funcName = self.getSourceFunctionName()
    body = """Function '%s' expected the argument: '%s' to be an instance of 
    a class derived from metaclass: '%s', but '%s' belongs to class: '%s' 
    which is derived from '%s'."""
    msg = body % (funcName, argName, actMcls, argName, actCls, actMcls)

    return '%s\n%s' % (header, self.justify(msg))



"""WorkToy - Wait A Minute! - MetaXcept
Metaclass creating the exception types"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from inspect import FrameInfo, stack, trace
from typing import Never, Any

from icecream import ic

from worktoy.worktoyclass import WorkToyClass
from worktoy.core import Function
from worktoy.metaclass import AbstractMetaClass, AbstractNameSpace
from worktoy.waitaminute import ExceptSpace

ic.configureOutput(includeContext=True)


class _MetaXcept(AbstractMetaClass):
  """WorkToy - Wait A Minute! - MetaXcept
  Metaclass creating the exception types"""

  @classmethod
  def __prepare__(mcls, *args, **kwargs) -> dict:
    name, bases = args[0], args[1]
    return AbstractNameSpace(name, bases, **kwargs)

  def __new__(mcls, *args, **kwargs) -> type:
    name = args[0]
    bases = args[1]
    nameSpace = args[2]
    out = AbstractMetaClass.__new__(mcls, name, bases, nameSpace, **kwargs)
    out.__name_space__ = nameSpace
    return out

  def __init__(cls, *args, **kwargs) -> None:
    name = args[0]
    bases = args[1]
    nameSpace = args[2]
    if isinstance(nameSpace, dict):
      AbstractMetaClass.__init__(cls, name, bases, nameSpace, **kwargs)
    else:
      raise TypeError

  def __call__(cls, *args, **kwargs) -> Exception:
    return AbstractMetaClass.__call__(cls, *args, **kwargs)

  def __eq__(cls, other: type) -> bool:
    """Other must be a type"""
    return True if cls is other else False


class MetaXcept(Exception, WorkToyClass, metaclass=_MetaXcept):
  """In between class exposing the metaclass."""

  def formatHeader(self, header: str) -> str:
    """Formats the given string to a cool header"""
    side = (77 - (len(header) + 6)) / 2
    sideLine = '~' * int(side)
    width = 73 - (2 * side) % 2
    topLine = '| %s |' % (int(width) * '¤')
    bottomLine = '| %s |' % (int(width) * '^')
    coolName = '| %s %s %s |' % (sideLine, header, sideLine)
    return '\n\n%s\n%s\n%s\n' % (topLine, coolName, bottomLine)

  def formatFuncName(self, func: Function) -> str:
    """Assigns name to function."""
    func = getattr(func, '__func__', func)
    funcQualName = getattr(func, '__qualname__', None)
    funcName = getattr(func, '__name__', None)
    funcPrintF = '%s' % func
    return self.maybe(funcQualName, funcName, funcPrintF)

  def formatTypeName(self, type_: type) -> str:
    """Formats the name of a type."""
    if not isinstance(type_, type):
      return self.formatTypeName(type(type_))
    return type_.__qualname__

  def pretty(self, arg: Any) -> str:
    """Formats argument arg."""
    if isinstance(arg, Function):
      return self.formatFuncName(arg)
    if isinstance(arg, type):
      return self.formatTypeName(arg)
    return str(arg)

  def __new__(cls, *args, **kwargs) -> MetaXcept:
    return Exception.__new__(cls)

  def __init__(self, *args, **kwargs) -> None:
    Exception.__init__(self, *args)
    WorkToyClass.__init__(self, *args, **kwargs)
    self._stack = stack()
    self._trace = trace()

  def getStack(self) -> list[FrameInfo]:
    """Getter-function for stack."""
    return self._stack

  def getReversedStack(self) -> list[FrameInfo]:
    """Getter-function for reversed stack."""
    _stack = [i for i in self.getStack()]
    out = []
    while _stack:
      out.append(_stack.pop())
    return out

  def getTrace(self) -> Any:
    """Getter-function for trace"""
    return self._trace

  def getFuncName(self) -> list[str]:
    """Getter-function for the latest function."""
    return [item[-1].function for item in self.getStack()]

  def getFuncQualName(self) -> Function:
    """Getter-function for the qualified names of the functions."""
    return [item.frame.f_code.co_qualname for item in self.getStack()]

  def getFuncs(self) -> Any:
    """Getter-function for functions. FFS"""
    raise NotImplementedError('Intentionally unavailable. Only names.')

  def getSourceFunctionStack(self) -> list:
    """Getter-function for the function stack."""
    out = []
    __exceptional_names__ = self.stringList("""MetaXcept, 
    AbstractMetaClass, _MetaXcept, __guard_validator__""")
    __exceptional_names__.append(self.__class__.__qualname__)
    for item in self.getFuncQualName():
      scope, func = [*item.split('.'), None, None][:2]
      if scope not in __exceptional_names__:
        out.append(item)
    return out

  def getSourceFunctionName(self) -> Any:
    """Getter-function for the function in which the error is raised.
    Unfortunately, only the name is available from the inspect module. """
    return self.getSourceFunctionStack()[-1]

  def getLocals(self) -> Any:
    """Getter-function for locals"""
    out = []
    for item in self.getStack():
      out.append(item.frame.f_locals)
    return out

  def __getattr__(self, key: str) -> object:
    cls = object.__getattribute__(self, '__class__')
    return object.__getattribute__(cls, key)

  def __eq__(self, other: object) -> bool:
    """Implementation of equal operator. Please note that this method
    relates to the instance, not to the class. The class (MetaXcept) has
    its own implementation.
    If other is a type, this method returns MetaXcept == other
    If other is a string:
      self.__str__() == other
    Otherwise:
      self.__str__() == other.__str__()"""
    if isinstance(other, type):
      return self.__class__ == other
    if isinstance(other, str):
      return other in [self.__str__(), self.__class__.__qualname__]
    return self == object.__str__(other)

  def handle(self, *args, **kwargs) -> Never:
    """Should be invoked after the exception is caught. By default,
    it raises itself. """
    raise self

  def __getattribute__(self, key: str) -> Any:
    val = object.__getattribute__(self, key)
    if key == '__module__' and isinstance(val, str):
      return '\n%s\n' % val
    return val

  def __str__(self) -> str:
    """String representation. The default returns the qualified name of
    the error with the new lines above and below."""
    return self.formatHeader(self.__class__.__qualname__)

  def combine(self, other: MetaXcept) -> str:
    """String representation"""
    selfHeader = MetaXcept.__str__(self)
    otherHeader = MetaXcept.__str__(other)
    selfName = self.__class__.__qualname__
    otherName = other.__class__.__qualname__
    title = MetaXcept.__qualname__
    headerContent = """%s encountered multiple errors!""" % title
    subHeaderContent = '%s and %s:' % (selfName, otherName)
    header = self.formatHeader(headerContent)
    subHeader = self.formatHeader(subHeaderContent)
    selfMessage = str(self).replace(selfHeader, '')
    otherMessage = str(other).replace(otherHeader, '')
    return '\n'.join([header, subHeader, selfMessage, otherMessage])

  # def collectInfo(self, ) -> list:
  #   """Collects details for each function step"""
  #   out = []
  #   for f in self.getReversedStack():
  #     out.append(self.collectStepInfo(f))
  #   return out
  #
  # def collectNames(self, ) -> list:
  #   """Collects function names."""
  #
  # def collectStepInfo(self, step: Any) -> dict:
  #   """Collects details for one step"""
  #   globalSnapshot = globals().items()
  #   globalSnapshot = {k: v for (k, v) in globalSnapshot}
  #   name = step.frame.f_code.co_qualname
  #   argNames = step.frame.f_code.co_qualname
  #   func = globalSnapshot.get(name)
  #   note = getattr(func, '__annotations__', None)
  #   if note is None:
  #     return {}
  #   notes = {}
  #   for arg in [*argNames, 'return']:
  #     notes |= {arg: note.get(arg, 'None')}
  #   return dict(name=name, func=func, notes=notes)
  #
  # def funcInfo(self, ) -> Any:
  #   """Function information"""
  #   out = []
  #   for info in self.collectInfo():
  #     if info.get('name', None) != '<module>':
  #       out.append(info)
  #   return out



"""WorkToy - Wait A Minute! - UnsupportedSubclassException
This exception should be raised when encountering a variable of correct
type, but of incorrect subclass."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.waitaminute import MetaXcept


class UnsupportedSubclassException(MetaXcept):
  """WorkToy - Wait A Minute! - UnsupportedSubclassException
  This exception should be raised when encountering a variable of correct
  type, but of incorrect subclass."""

  def __init__(self, name: str, exp: type, obj: object,
               *args, **kwargs) -> None:
    MetaXcept.__init__(self, *args, **kwargs)
    self._actualName = self.pretty(name)
    self._expectedParent = self.pretty(exp)
    self._actualVariable = self.pretty(obj)

  def __str__(self, ) -> str:
    header = MetaXcept.__str__(self)
    name = self._actualName
    exp = self._expectedParent
    obj = self._actualVariable
    msg = """Expected argument: '%s' to be an instance of a subclass of 
    '%s', but received '%s' of type: '%s' having the following 
    incompatible method resolution order:"""
    body = self.monoSpace(msg % (name, exp, obj, type(obj)))
    _mro = [arg for arg in type(obj).__mro__()]
    _mro = [self.pretty(arg) for arg in _mro]
    _mro = '<br>  '.join(_mro)
    body = '%s<br>%s' % (body, _mro)
    return '%s\n%s' % (header, self.justify(body))



"""WorkToy - SYM
Symbolic classes (Enum implementation)

Procedure for creation of symbolic classes.

The new class should inherit from a symbolic baseclass such as the 'SYM'
class which may be used directly or as a baseclass for a custom
implementation. In the class body of the new class, create the instances.
The 'SYM' class provides a way of creating instances. The default
implementation uses the 'auto' method. For example,

  class WeekDay(SYM):
    # Symbolic class representation of weekdays
    MONDAY=SYM.auto()
    TUESDAY=SYM.auto()
    ...

Note that the instances are named in upper case as is convention. The
worktoyclass
implementation is case-insensitive, but presents instances in upper case.
This means that:

 WeekDay.Wednesday is 'WeekDay.WEDNESDAY'

The SYM baseclass implements hashing allowing dictionaries to use
instances as keys. The SYM baseclass also implements '__eq__' and '__ne__'.

In the SyMeta metaclass, the SymSpace class is used to provide the
namespace object returned by the '__prepare__' method. The SymSpace class
relies on the SYM-subclass to decide if an entry in the class body denotes
an instance creation. The SymSpace dictionary splits the data to
instanceSpace and to namespace. The namespace data is then used in the
super call as part of the normal class creation process, and the
instanceSpace data is used in the __init__ method on the metaclass to
facilitate instance creation. Finally, the metaclass implements the
__call__ method to restrict instance creation to those instances defined
in the instanceSpace.

A custom implementation of the SYM class must provide the following to be
recognized by the rest of the SYM module:
  - Identification. The custom implementation should explicitly define
  itself as a symbolic baseclass by setting:
    '__symbolic_baseclass__ = True'
  If no baseclass has this variable set, the SymSpace defaults to using
  the default SYM class, even if an intended symbolic baseclass is
  present. This allows symbolic classes to inherit from other classes.

  - Validation. The custom implementation must provide a method called
  'validateInstance' mapping '[str, object] -> bool' which is used by the
  SymSpace instance to determine when an entry in the class body is to
  mean instance of the symbolic class.

  - Differentiation. The custom implementation must set a unique value on
  each instance. By default, this value is an integer assigned in the order
  received in the class body beginning at '0'. Custom implementations are
  free to use any data type and reimplement the '__eq__' method as
  appropriate.

  - Instantiation. The custom implementation is responsible for creating
  the instance given a key, value pair. The default implementation creates
  a worktoyclass instance of the new class and an instance of itself that
  wraps
  it. Please note that the default implementation requires the new class
  to support instance creation with: 'instance = NewClass()'. Otherwise,
  necessary arguments must be provided in the class body, for example,
    * 'instance = SYM.auto(420, 69)'
    The above will result in the 'instance' wrapping: 'NewClass(420, 69)'

  - Parsing (OPTIONAL). If provided, the 'parseInstance' method should take
  '*args, **kwargs' and return the instance matching or raise a NameError.
  The default implementation allows for the new class to be callable. By
  default, the instance value ('int') or the instance ('name') may be used
  to retrieve the desired instance by calling the class. For example:
    WeekDay.Tuesday == WeekDay('tuesday')  # case-insensitive
  If the parsing method is not defined, or the new class itself implements
  '__call__' the above will not be available.

  - Decoration (OPTIONAL). The default implementation provides the
  following enhancements of the new class:
    * Iteration:
      for item in 'Weekday':
        - 'Weekday.MONDAY'  # the wrapped instance
        - 'Weekday.TUESDAY'
        ...
      The above is achieved with '__iter__', '__next__' and '__len__'
      implemented as class methods. The '__len__' is taken to mean the
      number of instances on the class.
    * Dictionary-like behaviour:
      - Implementation of 'keys', 'values' and 'items'.
      - Implementation of '__getitem__':
        * 'Weekday[0]' -> 'Weekday.MONDAY'  # 'int' refers to value
        * 'Weekday['monday']' -> 'Weekday.MONDAY'  # 'str' refers to name
      - Implementation of binary arithmetic operations including cases of
        reflected operands.
        * 'Weekday[2] + 3' = 'Weekday.SATURDAY'  # SYM.__add__(self, other)
        * '3 + Weekday[2] = 5'  # SYM.__radd__(self, other)
        * 'Weekday[2] += 1 -> Weekday[3]' # SYM.__iadd__(self, other)
        The arithmetic implementation provides integer like behaviour. So
        when an instance appears as the 'other' operand, it behaves like an
        integer of the same value as itself.
        Instances use modular arithmetic such that:
          - 'Weekday.SATURDAY' + 2 -> 'Weekday.MONDAY'
        In general with n = len(SYM):
          -  'SYM.INSTANCE[a] + SYM.INSTANCE[b] = SYM.INSTANCE[(a + b) % n]'
  The default implementation provides parsing such that:
    Weekday(4) -> 'Weekday.FRIDAY'
    Weekday('friday') -> 'Weekday.FRIDAY'  # Case insensitive by default.
"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from ._sym import SYM
from ._symspace import SymSpace
from ._symeta import SyMeta, BaseSym



"""WorkToy - Wait A Minute! - MissingArgumentException
This exception should be raised when a function is missing a required
argument."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.waitaminute import MetaXcept


class MissingArgumentException(MetaXcept):
  """WorkToy - Wait A Minute! - UnexpectedStateError
  Invoked when a function is called before it is ready."""

  def __init__(self, missingArgument: str, *args, **kwargs) -> None:
    MetaXcept.__init__(self, missingArgument, *args, **kwargs)
    self._missingArgument = self.pretty(missingArgument)

  def __str__(self, ) -> str:
    header = MetaXcept.__str__(self)
    misArg = self._missingArgument
    funcName = self.getSourceFunctionName()
    msg = """Variable: '%s' is required for function: '%s' is missing!"""
    body = msg % (misArg, funcName)
    return '%s\n%s' % (header, self.justify(body))



"""WorkToy - SYM - SyMeta
Metaclass implementing Symbolic classes."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, TYPE_CHECKING

from icecream import ic

from worktoy.worktoyclass import WorkToyClass
from worktoy.core import Bases
from worktoy.metaclass import AbstractMetaClass, MetaMetaClass
from worktoy.sym import SYM, SymSpace

if TYPE_CHECKING:
  pass

ic.configureOutput(includeContext=True)


class SyMetaMeta(MetaMetaClass):
  """Applies subclass check on the metaclass"""

  def __subclasscheck__(cls, subclass):
    names = ['__symbolic_class__', '__symbolic_baseclass__']

    for name in names:
      if hasattr(subclass, name):
        return True
    return False

  def __instancecheck__(cls, instance):
    return True if cls.__subclasscheck__(instance.__class__) else False


class SyMeta(AbstractMetaClass, metaclass=SyMetaMeta):
  """WorkToy - SYM - SyMeta
  Metaclass implementing Symbolic classes. """

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> dict:
    """Returning special namespace class."""
    return SymSpace(name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Bases, data: SymSpace, **kw) -> type:
    nameSpace = data.getNameSpace()
    instanceSpace = data.getInstanceSpace()
    instanceSpace = {k.lower(): v for (k, v) in instanceSpace.items()}
    out = AbstractMetaClass.__new__(mcls, name, bases, nameSpace, )
    setattr(out, '__instance_space__', instanceSpace)
    setattr(out, '__symbolic_instances__', {})
    setattr(out, '__symbolic_class__', True)
    return out

  def __init__(cls, *args, **kwargs) -> None:
    type.__init__(cls, *args, **kwargs)
    symNames = kwargs.get('symNames', {})
    for key, val in symNames.items():
      sym = SYM()
      sym.__set_name__(cls, key.lower())
      setattr(sym, 'value', val)

  def __call__(cls, *args, **kwargs) -> Any:
    if kwargs.get('__instance_creation__', False):
      self = type.__call__(cls, *args, **kwargs)
      existing = getattr(cls, '__symbolic_instances__', [])
      setattr(cls, '__symbolic_instances__', [*existing, self])
      return self
    index, key = None, None
    for arg in args:
      if isinstance(arg, int) and index is None:
        index = arg
      if isinstance(arg, str) and key is None:
        key = arg
    if key is not None:
      keyVal = cls.getInstanceAtKey(key.lower(), _allowError=True)
      if keyVal is not None:
        return keyVal
    indexVal = cls.getInstanceAtIndex(index, _allowError=True)
    if indexVal is not None:
      return indexVal

  def getInstances(cls, ) -> list[SYM]:
    """Getter-function for symbolic instances."""
    syms = [v for (k, v) in cls.__dict__.items() if isinstance(v, SYM)]
    return [sym for sym in syms]

  def getInstanceAtIndex(cls, index: int, **kwargs) -> SYM:
    """Getter-function for the item at the given index."""
    existing = getattr(cls, '__symbolic_instances__', None)
    for (i, item) in enumerate(existing):
      if i == index:
        return item

  def getInstanceAtKey(cls, name: str, **kwargs) -> SYM:
    """Getter-function for the item at the given key."""
    existing = getattr(cls, '__symbolic_instances__', [])
    for (i, item) in enumerate(existing):
      if item.getName().lower() == name.lower():
        return item
    raise KeyError(name)

  def __instancecheck__(cls, instance: Any, ) -> bool:
    test = getattr(instance, '__symbolic_class__', None)
    return True if test else False

  def __repr__(cls) -> str:
    """Code Representation"""
    return '%s.%s()' % (cls.__class__.__qualname__, cls.__qualname__)

  def __str__(cls) -> str:
    """String Representation"""
    clsName = cls.__qualname__
    mclsName = cls.__class__.__qualname__
    return 'Symbolic Class: %s.%s' % (mclsName, clsName)

  def __getattr__(cls, key: str) -> Any:
    try:
      return object.__getattribute__(cls, key.lower())
    except AttributeError as e:
      raise e

  def __iter__(cls, ) -> type:
    setattr(cls, '__current_index__', 0)
    return cls

  def __next__(cls) -> Any:
    ind = getattr(cls, '__current_index__', None) + 1
    if ind is None:
      raise TypeError
    instances = cls.getInstances()
    if ind > len(instances):
      raise StopIteration
    setattr(cls, '__current_index__', ind)
    return instances[ind - 1]


class BaseSym(WorkToyClass, metaclass=SyMeta):
  """In between class exposing the metaclass."""

  def __init_subclass__(cls, **kwargs) -> None:
    setattr(cls, '__symbolic_class__', True)

  def __call__(self, *args, **kwargs) -> Any:
    return self.__class__.__call__(*args, **kwargs)

  def __str__(self, ) -> str:
    key = None
    for key, val in self.__class__.__dict__.items():
      if isinstance(val, SYM):
        if val.innerInstance is self:
          break

    return '%s.%s' % (self.__class__.__qualname__, key)



"""WorkToy - Core
This package provides core types."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from ._coretypes import Map, Keys, Values, Items, Bases, Type
from ._coretypes import Function, Method, WrapperDescriptor
from ._coretypes import WrapperMethod, BuiltinFunction, Functional
from ._coretypes import FunctionTuple, FunctionList
from ._coretypes import ARGS, KWARGS, RESULT, CALL
from ._math import PI, factorial, EXP1



"""Test of GuardClass"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations
from unittest import TestCase

import unittest
from worktoy.worktoyclass import WorkToyClass


class TestDefaultClass(unittest.TestCase):

  def setUp(self):
    self.default = WorkToyClass()

  def test_monoSpace_with_default_newline(self):
    text = 'Hello   World'
    result = self.default.monoSpace(text)
    self.assertEqual(result, 'Hello World')

  def test_monoSpace_with_custom_newline(self):
    text = 'Hello<br>World'
    result = self.default.monoSpace(text, '<br>')
    self.assertEqual(result, 'Hello\nWorld')

  def test_stringList_with_default_separator(self):
    result = self.default.stringList("""one, two, three, four""")
    self.assertEqual(result, ['one', 'two', 'three', 'four'])

  def test_stringList_with_custom_separator(self):
    result = self.default.stringList('one | two | three | four',
                                     separator=' | ')
    self.assertEqual(result, ['one', 'two', 'three', 'four'])

  def test_justify(self):
    text = self.default.monoSpace("""This is a long text that needs to be 
      justified to fit within a certain character limit.""")
    text = self.default.monoSpace(text)
    result = self.default.justify(text, 50)
    expect = self.default.monoSpace("""This is a long text that needs to 
      be justified<br>to fit within a certain character limit.""")
    self.assertEqual(result, expect)

  def test_extractBetween(self):
    text = 'a(b(c)d)e'
    result = self.default.extractBetween(text, '(', ')')
    self.assertEqual(result, ['b(c'])

  def test_trimWhitespace(self):
    text = '  Some text with leading and trailing spaces   '
    result = self.default.trimWhitespace(text)
    self.assertEqual(result, 'Some text with leading and trailing spaces')

  def test_typeName_with_type(self):
    result = self.default.typeName(str)
    self.assertEqual(result, 'str')


if __name__ == '__main__':
  unittest.main()



"""WorkToy - Core - Type collection
This module provides shorthands for many commonly used types."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Union

Sample = type('Sample', (), dict(clsMethod=classmethod(lambda cls: cls)))
Map = Union[type(type('_', (), {}).__dict__), dict]
Keys = type({}.keys())
Values = type({}.values())
Items = type({}.items())
Bases = tuple[type, ...]
Type = type(type('_', (), {}))
Function = (type(getattr(type('_', (), {'_': lambda self: self}), '_')))
Method = (type(getattr(type('_', (), {'_': lambda self: self})(), '_')))
WrapperDescriptor = type(type('_', (), {}).__init__)
WrapperMethod = type(type('_', (), {}).__call__)
BuiltinFunction = type(print)
Functional = Union[WrapperDescriptor, WrapperMethod, Function, Method]
FunctionTuple = (Function, Method, WrapperDescriptor, WrapperMethod,
                 BuiltinFunction)
FunctionList = [*FunctionTuple, ]
Functions = Union[
  Function, Method, WrapperDescriptor, WrapperMethod, BuiltinFunction]

ARGS = tuple[object, ...]
KWARGS = dict[str, object]
RESULT = tuple[tuple[object, ...], dict[str, object]]
CALL = tuple[Function, ARGS, KWARGS]



"""WorkToy - Descriptors
Implementation of descriptors as descriptors."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from ._field import Field
from ._attribute import AbstractAttribute
from ._int_attribute import IntAttribute
from ._float_attribute import FloatAttribute
from ._str_attribute import StrAttribute
from ._flag import Flag



"""WorkToy - Wait A Minute! - UnexpectedEventException
Raised when receiving a QEvent of the wrong Type."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.waitaminute import MetaXcept
from PySide6.QtCore import QEvent

TYPE = QEvent.Type


class UnexpectedEventException(MetaXcept):
  """WorkToy - Wait A Minute! - UnexpectedEventException
  Raised when receiving a QEvent of the wrong Type."""

  def __init__(self, expEvent: TYPE, actEvent: QEvent, argName: str,
               *args, **kwargs) -> None:
    MetaXcept.__init__(self, *args, **kwargs)
    self._expectedType = self.pretty(expEvent)
    self._actualEvent = self.pretty(actEvent)
    self._actualType = self.pretty(actEvent.Type)
    self._argName = self.pretty(argName)

  def __str__(self, ) -> str:
    header = MetaXcept.__str__(self)
    expType = self._expectedType
    actType = self._actualType
    actVal = self._actualValue
    argName = self._argName
    funcName = self.getSourceFunctionName()
    body = """Function '%s' expected event '%s' to be of event type: '%s', 
    but received event of type: '%s'. """
    msg = body % (funcName, argName, expType, actType)
    return '%s\n%s' % (header, self.justify(msg))



"""WorkToy - Descriptors - Flag
Implementation of boolean valued descriptor."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.descriptors import AbstractAttribute


class Flag(AbstractAttribute):
  """WorkToy - Descriptors - Flag
  Implementation of boolean valued descriptor."""

  @classmethod
  def getClassDefaultValue(cls) -> int:
    """Getter-function for default value."""
    return False

  def getType(self) -> type:
    """Getter-function for value type."""
    return bool



"""WorkToy - MetaClasses - MetaNameSpace
Provides a metaclass for creation of namespaces. """
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from icecream import ic

from worktoy.core import Bases
from worktoy.metaclass import AbstractMetaClass

ic.configureOutput(includeContext=True)


class MetaNameSpace(AbstractMetaClass):
  """WorkToy - MetaClasses - MetaNameSpace
  Provides a metaclass for creation of namespaces. """

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases) -> dict:
    return {}

  def __new__(mcls,
              name: str,
              bases: Bases,
              nameSpace: dict,
              **kw) -> type:
    return AbstractMetaClass.__new__(mcls, name, bases, nameSpace, **kw)

  def __init__(cls, name: str, bases: Bases, nameSpace: dict, **kw) -> None:
    AbstractMetaClass.__init__(cls, name, bases, nameSpace, **kw)
    setattr(cls, '__meta_name__', name)

  def __call__(cls, *args, **kwargs) -> object:
    return AbstractMetaClass.__call__(cls, *args, **kwargs)



"""WorkToy - Wait A Minute! - MissingKeyHandleError
When implementing a custom metaclass, a particularly powerful tool is to
implement custom namespace objects by reimplementing the __prepare__
method. The object returned by the __prepare__ method is expected to
behave like a builtin dictionary instance in a number of situations. The
custom exception defined here concerns how the object handles missing keys.
The namespace objected is expected to raise a KeyError with the name of
the given key. If the object fails to raise such a KeyError, highly
undefined behaviour will result. The system will not raise any informative
error message.

Consider the following example:

  class SusNameSpace(dict):
    #A custom namespace class.#

    def __init__(self, name, bases, *args, **kwargs) -> None:
      dict.__init__(self)
      self._className = name
      self._classBases = bases

    def __getitem__(self, key: str, ) -> object:
      if key in self.keys():
        return dict.__getitem__(self, key)


  class MyMetaClass(type):
    #A Typical metaclass.#

    @classmethod
    def __prepare__(mcls, name, bases, **kwargs):
      return SusNameSpace(name, bases)


  class MyClass(metaclass=MyMetaClass):
    #A class derived from the metaclass#

    @staticmethod
    def myMethod():
      #A static method. #
      pass

The above results in the following error message:

  Traceback (most recent call last):
    File "...", line 14, in <module>
      from ... import MyClass
    File "...", line 28, in <module>

      class MyClass(metaclass=MyMetaClass):
    File "...", line 31, in MyClass
      @staticmethod
       ^^^^^^^^^^^^
  TypeError: 'NoneType' object is not callable

  Process finished with exit code 1

The error encountered is because the SusNameSpace instance did not raise:
KeyError('staticmethod'). Changing the __getitem__ method to default
dictionary behaviour, but including a print statement, reveals what is
happening:

  ...
  def __getitem__(self, key: str, ) -> object:
    try:
      return dict.__getitem__(self, key)
    except Exception as e:
      print(key)
      raise e
  ...

Now, the console outputs the following:
'''
  __name__
  staticmethod

  Process finished with exit code 0
'''

The metaclass system implemented in the WorkToy package includes
validation of the namespace objects returned by the __prepare__ method.
The MissingKeyHandleError is raised when a namespace 'objects' fails to
raise KeyError against missing keys in the __getitem__.

Custom metaclasses are powerful, but have such odd quirks. """
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.worktoyclass import WorkToyClass
from worktoy.core import Function
from worktoy.waitaminute import MetaXcept


class MissingKeyHandleError(MetaXcept):
  """...namespace objects fails to raise KeyError against missing keys in
  the __getitem__"""

  monoSpace = WorkToyClass().monoSpace

  def __init__(self, prepare: Function, **kwargs) -> None:
    mcls = kwargs.get('metaclass')
    MetaXcept.__init__(self, prepare, mcls)
    self._prepare = prepare
    self._metaClass = mcls

  def __str__(self) -> str:
    title = ('<br>CRITICAL!<br>%s<br>FATAL!<br>' %
             self.__class__.__qualname__)
    header = """The namespace produced by the given function %s during the 
    class creation implemented by metaclass %s, """
    header = header % (str(self._prepare), str(self._metaClass))
    body = """returned unexpected behaviour against missing keys in the 
    __getitem__ method!<br>Custom classes used as namespaces must raise: 
    "KeyError('key')" where 'key' is the string representing a missing 
    'key'."""
    out = ''.join([title, header, body])
    return self.monoSpace(out)



"""WorkToy - Core - Math
Module providing math tools."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

factorial = lambda n: 1 if n < 2 else n * (n - 1)
PI = 3.141592653589793
EXP1 = [1 / factorial(i) for i in range(16)]



"""WorkToy - Core
The core module provides the chain of default classes:
  'CoreClass'
  'GuardClass'
  'WorkToyClass'
The final class should be called DefaultClass."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from ._core_class import CoreClass
from ._guard_class import GuardClass
from ._worktoy_class import WorkToyClass

if __name__ != '__main__':
  WorkToyClass.__core_instance__ = WorkToyClass()



"""WorkToy - Base - GuardClass
Provides guard methods to the DefaultClass"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.worktoyclass import CoreClass

Function = (type(getattr(type('_', (), {'_': lambda self: self}), '_')))


class GuardClass(CoreClass):
  """WorkToy - Base - GuardClass
  Provides guard methods to the DefaultClass"""

  def __init__(self, *args, **kwargs) -> None:
    CoreClass.__init__(self, *args, **kwargs)

  def noneGuard(self, obj: object, varName: str = None) -> Any:
    """Raises UnavailableNameException if the given object is not None. """
    if obj is not None:
      from worktoy.waitaminute import UnavailableNameException
      raise UnavailableNameException(self.maybe(varName, 'obj'), obj)

  def someGuard(self, obj: Any, varName: str = None) -> Any:
    """Raises error if given object is None."""
    if obj is None:
      from worktoy.waitaminute import MissingArgumentException
      raise MissingArgumentException(self.maybe(varName, 'obj'))
    return obj

  def overRideGuard(self, obj: Any, varName: str, newValue: Any) -> Any:
    """Raises error if given object is not None. """
    if obj is not None:
      from worktoy.waitaminute import UnavailableNameException
      name = varName
      oldVal = obj
      newVal = newValue
      raise UnavailableNameException(name, oldVal, newVal)

  def functionGuard(self, func: Function, name: str = None) -> Function:
    """Raises error if given object is not a function."""
    argName = self.maybe(name, 'func')
    if not callable(self.someGuard(func, argName)):
      from worktoy.waitaminute import TypeSupportError
      expectedType = Function
      actualValue = func
      raise TypeSupportError(expectedType, actualValue, argName)
    return func

  def intGuard(self, integer: int, name: str = None) -> int:
    """Raises error if given object is None or not an integer."""
    argName = self.maybe(name, 'integer')
    if not isinstance(self.someGuard(integer), int):
      from worktoy.waitaminute import TypeSupportError
      expectedType = int
      actualValue = integer
      raise TypeSupportError(expectedType, actualValue, argName)
    return integer

  def floatGuard(self, value: int, name: str = None) -> int:
    """Raises error if given object is None or not a float."""
    argName = self.maybe(name, 'float')
    if not isinstance(self.someGuard(value), (int, float)):
      from worktoy.waitaminute import TypeSupportError
      expectedType = float
      actualValue = value
      raise TypeSupportError(expectedType, actualValue, argName)
    return value

  def strGuard(self, value: str, name: str = None) -> str:
    """Raises error if given object is None or not a string."""
    argName = self.maybe(name, 'text')
    if not isinstance(self.someGuard(value), str):
      from worktoy.waitaminute import TypeSupportError
      expectedType = str
      actualValue = value
      raise TypeSupportError(expectedType, actualValue, argName)
    return value



"""WorkToy - WorkToyClass - WorkToyClass
This class provides utility as static methods.
Classes in the package subclass this class."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from warnings import warn

from worktoy.worktoyclass import GuardClass


class WorkToyClass(GuardClass):
  """WorkToy - DefaultClass
  This class provides utility as static methods.
  Classes in the package subclass this class.
  """

  __core_instance__ = None

  def __init__(self, *args, **kwargs) -> None:
    GuardClass.__init__(self, *args, **kwargs)

  def monoSpace(self, text: str, newLine: str = None) -> str:
    """Convert text to monospaced format with consistent spacing and line
    breaks.

    Args:
      text (str): The input text to be modified.
      newLine (str, optional):
        The string representing the line break. If not provided,
        the default value '<br>' is used. Defaults to None.

    Returns:
      str: The modified text with consistent spacing and line breaks.

    Raises:
      None

    Examples:
      >>> self.monoSpace('Hello   World')
      'Hello World'
      >>> self.monoSpace('Hello<br>World', '<br>')
      'Hello\nWorld'

    The `monoSpace` function takes a string `text` and an optional string
    `newLine`, and returns a modified version of the input text with
    consistent
    spacing and line breaks. If the `newLine` argument is not provided, the
    default value '<br>' is used as the line break string.

    The function performs the following steps:
    1. Replaces all occurrences of '\n' and '\r' characters with a space '
    ' in
       the input `text`.
    2. Repeatedly replaces multiple consecutive spaces with a single space
       until no more consecutive spaces are found in the `text`.
    3. Replace the `newLine` string (or the default '<br>' if not provided)
       with a line break '\n' character in the modified `text`.
    4. Returns the modified `text`.

    Note:
    - The `newLine` string is treated as a literal string, so make sure to
      provide the exact string to be replaced as the line break.
    - The `newLine` string is case-sensitive. If the provided `newLine`
    string
      does not match the exact case in the input `text`, it will not be
      replaced with a line break."""
    newLine = str(self.maybe(newLine, '<br>'))
    text = text.replace('\n', ' ').replace('\r', ' ')
    while '  ' in text:
      text = text.replace('  ', ' ')
    return text.replace(newLine, '\n')

  def stringList(self, *args, **kwargs) -> list[str]:
    """
    The stringList function provides an easier way to write lists of strings.
    Instead of wrapping each item in ticks, write on long string with
    consistent separators, and stringList will convert it to a list of
    strings.
    Instead of: numbers = ['one', 'two', 'three', 'four']
    Use stringList: numbers = stringList('one, two, three, four')
    Please note that all white space around each separator will be removed.
    For example: ', ' and ',' will produce the same outcome when used as
    separators on the same text.
    #  MIT Licence
    #  Copyright (c) 2023 Asger Jon Vistisen"""

    strArgs = self.maybeTypes(str, *args, pad=3)
    sourceKeys = ['source', 'src', 'txt']
    sourceKwarg = self.maybeKey(str, *sourceKeys, **kwargs)
    separatorKeys = ['separator', 'splitHere']
    separatorKwarg = self.maybeKey(str, *separatorKeys, **kwargs)
    sourceArg, separatorArg, ignoreArg = strArgs
    sourceDefault, separatorDefault = None, ', '
    source = self.maybe(sourceKwarg, sourceArg, sourceDefault)
    separator = self.maybe(separatorKwarg,
                           separatorArg,
                           separatorDefault, )
    if source is None:
      msg = 'stringList received no string!'
      raise ValueError(msg)
    if not isinstance(source, str):
      raise TypeError
    out = source.split(str(separator))
    out = [self.monoSpace(arg) for arg in out]
    return [self.trimWhitespace(arg) for arg in out]

  def justify(self,
              text: str,
              charLimit: int = None,
              *spaceLike: str) -> str:
    """The justify function works similarly to monoSpace, except for
    ignoring all new-line indicators and instead splits the string into
    lines separated by new-line at given intervals. """
    if not text:
      return ''
    text = text.replace(' %', '¤¤')
    text = self.monoSpace(text)
    charLimit = int(self.maybe(charLimit, 77))
    spaceLike = [' ', *spaceLike]
    allWords = [text.split(space) for space in spaceLike]
    wordList = []
    for words in allWords:
      wordList = [*wordList, *words]
    words = wordList
    lines = []
    line = []
    for (i, word) in enumerate(words):
      if sum([len(w) for w in line]) + len(word) + len(line) + 1 < charLimit:
        line.append(word)
      else:
        lines.append(line)
        line = [word, ]
    lines.append(line)
    lineList = []
    for line in lines:
      lineList.append(' '.join(line))
    out = '\n'.join(lineList)
    out = out.replace('¤¤', ' %')
    return out

  def extractBetween(self, text: str, before: str, after: str) -> list:
    """
    Extracts substrings from the text that are enclosed between the specified
    characters, before, and after. If before and after are the same
    character,
    the function will correctly handle this case and extract the text between
    each pair of those characters.

    :param text: The text to search within can be a multiline string.
    :param before: The character or substring that precedes the text to be
                   extracted. If multiline, can be a newline character.
    :param after: The character or substring that follows the text to be
                  extracted. If multiline, can be a newline character.
    :return: A list of substrings found between the specified characters.
             If no such substrings are found, it will return an empty list.
    :example: extractBetween("a(b(c)d)e", '(', '(') returns ['b(c']
              extractBetween("Line 1\nLine 2\nLine 3", '\n', '\n') returns
              ['Line 2']
    """
    result = []
    start = text.find(before)
    while start != -1:
      if before == after:
        end = text.find(after, start + 2)
      else:
        end = text.find(after, start + 1)
      if end != -1:
        result.append(text[start + 1:end])
      start = text.find(before, end + 1)
    return result

  def trimWhitespace(self, text: str) -> str:
    """
    Removes the leading and trailing whitespace from the input text.

    Args:
      text (str): the input text string.

    Returns:
      str: the text string without leading and trailing whitespace.
    """
    return text.strip()

  def typeName(self, cls: object) -> str:
    """Returns the name of the given type. If the argument is not a type,
    the type of the argument is used instead."""
    if not isinstance(cls, type):
      msg = """Expected argument to be of type 'type', but received: %s"""
      warn(msg % type(cls))
    if cls is None:
      return 'NoneClass'
    qualName = getattr(cls, '__qualname__', None)
    name = getattr(cls, '__name__', None)
    strRep = '%s' % (cls)
    out = self.maybe(qualName, name, strRep)
    if isinstance(out, str):
      return out



"""WorkToy - Wait A Minute! - ErrorNameSpace
Special namespace class for use with the metaclass."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.core import Keys, Items, Values, Bases
from worktoy.metaclass import AbstractNameSpace


class ExceptSpace(AbstractNameSpace):
  """WorkToy - Wait A Minute! - ErrorNameSpace
  Special namespace class for use with the metaclass."""

  def __init__(self, name: str, bases: tuple[type], **kwargs) -> None:
    AbstractNameSpace.__init__(self, name, bases, **kwargs)

  def getModuleName(self) -> str:
    """Getter-function for the module name"""
    return 'WorkToy - Wait A Minute!'



"""WorkToy - SYM - SYM
The basic class implementing Enum like behaviour."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Never

from icecream import ic

from worktoy.worktoyclass import WorkToyClass

ic.configureOutput(includeContext=True)


class SYM(WorkToyClass):
  """WorkToy - SYM - AUTO
  Special class denoting an instance on the class body."""
  __symbolic_baseclass__ = True
  __decorated_classes__ = {}

  def __init__(self, *args, **kwargs) -> None:
    WorkToyClass.__init__(self, *args, **kwargs)
    self._value = None
    self._name = None
    self._owner = None
    self.innerClass = None
    self.innerInstance = None
    setattr(self, '__symbolic_instance__', True)

  @classmethod
  def auto(cls, *args, **kwargs) -> SYM:
    """Used to define a new instance"""
    return cls(*args, **kwargs)

  def getName(self) -> str:
    """Getter-function for name"""
    return self._name

  def __set_name__(self, cls: type, name: str) -> None:
    self._name = name
    self._owner = cls
    existing = self.__class__.__decorated_classes__.get(cls, [])
    self.__class__.__decorated_classes__[cls] = [*existing, cls]
    args = self.getArgs()
    kwargs = dict(__instance_creation__=True, **self.getKwargs())
    newInstance = cls(*args, **kwargs)
    setattr(self, 'innerClass', cls)
    setattr(self, 'innerInstance', newInstance)
    existing = getattr(self._owner, '__symbolic_instances__', [])
    setattr(self, 'value', len(existing))
    setattr(self._owner, '__symbolic_instances__', [*existing, self])

  def __get__(self, obj: object, cls: type) -> Any:
    try:
      return self
    except AttributeError as e:
      raise e

  def __set__(self, obj: object, newValue: Any) -> Never:
    if isinstance(newValue, SYM):
      innerInstance = getattr(newValue, 'innerInstance', None)
      if innerInstance.__class__ == obj.__class__:
        obj = innerInstance

  def __delete__(self, *_) -> Never:
    raise NotImplementedError

  def __getattr__(self, key: str) -> Any:
    try:
      return object.__getattribute__(self, key.lower())
    except AttributeError as e:
      raise e

  def __str__(self, ) -> str:
    if self.innerClass:
      clsName = self.innerClass.__qualname__
      insName = self.innerInstance.__name__
      return 'SYM wrapper on \'%s\' of %s' % (self._name, clsName)
    return 'SYM before wrapping'

  def __repr__(self, ) -> str:
    clsName = '' if self.innerClass is None else self.innerClass.__qualname__
    return 'SYM(%s.%s)' % (clsName, self._name)



"""WorkToy - Descriptors - IntAttribute
Field implementation of integer valued descriptor."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.descriptors import AbstractAttribute


class IntAttribute(AbstractAttribute):
  """WorkToy - Fields - IntAttribute
  Field implementation of integer valued descriptor."""

  @classmethod
  def getClassDefaultValue(cls) -> int:
    """Getter-function for default value."""
    return 0

  def getType(self) -> type:
    """Getter-function for value type."""
    return int

  def __init__(self, value: int = None, *args, **kwargs) -> None:
    AbstractAttribute.__init__(self, value, *args, **kwargs)



"""WorkSide - Wait A Minute! - UnavailableName
Exception raised when a name is already occupied."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.waitaminute import MetaXcept


class UnavailableNameException(MetaXcept):
  """WorkSide - Wait A Minute! - UnavailableName
  Exception raised when a name is already occupied."""

  def __init__(self, name: str, oldVal: Any, newVal: Any = None,
               *args, **kwargs) -> None:
    self._name = name
    self._oldVal = oldVal
    self._newVal = newVal

  def __full_str__(self) -> str:
    """Message issued when receiving the new value."""
    header = MetaXcept.__str__(self)
    func = self.getFuncQualName()
    newVal = self._newVal
    name = self._name
    oldVal = self._oldVal
    msg = """Function '%s' attempted to assign new value: '%s' to name: 
    '%s' which is already populated with: '%s'!"""
    body = msg % (func, newVal, name, oldVal)
    return '%s\n%s' % (header, self.justify(body))

  def __name_val_str__(self) -> str:
    """Message issue when receiving only the name and existing value."""
    header = MetaXcept.__str__(self)
    name = self._name
    oldVal = self._oldVal
    func = self.getFuncQualName()
    msg = """Function '%s' expected variable at name '%s' to be 'None', 
    but found value: '%s'. """
    body = msg % (func, name, oldVal)
    return '%s\n%s' % (header, self.justify(body))

  def __str__(self, ) -> str:
    if self._newVal is not None:
      return self.__full_str__()
    return self.__name_val_str__()



"""WorkToy - MetaWork
This module provides the metaclass and namespace classes for the remaining
package."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

import worktoy.core as core
import worktoy.worktoyclass as worktoyclass
import worktoy.metaclass as metaclass
import worktoy.descriptors as descriptors
import worktoy.sym as sym
import worktoy.waitaminute as waitaminute



"""WorkToy - Wait A Minute! - NameSpaceError
General exception relating to namespaces in metaclass implementations."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.waitaminute import MetaXcept


class NameSpaceError(MetaXcept):
  """WorkToy - Wait A Minute! - NameSpaceError
  General exception relating to namespaces in metaclass implementations."""

  def __init__(self, msg: str, *args, **kwargs) -> None:
    MetaXcept.__init__(self, msg, *args, **kwargs)
    self._msg = msg

  def __str__(self) -> str:
    return self._msg



"""WorkToy - Wait A Minute! - RecursiveCreateGetError
Raised when a getter function calls a creator function a second time. """
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.core import Function
from worktoy.waitaminute import MetaXcept


class RecursiveCreateGetError(MetaXcept):
  """WorkToy - Wait A Minute! - RecursiveCreateGetError
  Raised when a getter function calls a creator function a second time. """

  def __init__(self, creator: Function, varType, varName: str,
               *args, **kwargs) -> None:
    MetaXcept.__init__(self, creator, varType, varName, *args, **kwargs)
    self._creator = self.pretty(creator)
    self._varType = self.pretty(varType)
    self._varName = self.pretty(varName)

  def __str__(self) -> str:
    header = MetaXcept.__str__(self)
    creator = self._creator
    varType, varName = self._varType, self._varName
    getter = self.getSourceFunctionName()
    msg = """Recursive behaviour detected between creator function: '%s' and 
    getter function '%s' for the variable: '%s: %s'."""
    msg = msg % (creator, getter, varName, varType)
    return '%s\n%s' % (header, self.justify(msg))



"""Test of CoreClass"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations
from unittest import TestCase

from worktoy.worktoyclass import CoreClass


class TestCoreClass(TestCase):
  """Test of CoreClass"""

  def test_maybe(self):
    """Test the maybe method."""
    coreObj = CoreClass()
    # Test with different types of arguments
    self.assertEqual(coreObj.maybe(None, 1, 2), 1)
    self.assertEqual(coreObj.maybe('a', None, 'b'), 'a')
    self.assertEqual(coreObj.maybe(None, None, 2.5), 2.5)
    # Test with all None values
    self.assertIsNone(coreObj.maybe(None, None, None))
    # Test with no arguments
    self.assertIsNone(coreObj.maybe())

  def test_maybeType(self):
    """Test the maybeType method."""
    coreObj = CoreClass()
    # Test with different types of arguments
    self.assertEqual(coreObj.maybeType(int, 'a', 1, 2.5), 1)
    self.assertEqual(coreObj.maybeType(str, 'a', 1, 'b'), 'a')
    self.assertEqual(coreObj.maybeType(float, None, 1, 2.5), 2.5)
    # Test with all incompatible values
    self.assertIsNone(coreObj.maybeType(int, 'a', 'b'))
    # Test with no arguments
    self.assertIsNone(coreObj.maybeType(int))

  def test_maybeTypes(self):
    """Test the maybeTypes method."""
    coreObj = CoreClass()
    # Test with different types of arguments and the specified types
    self.assertEqual(coreObj.maybeTypes(int, 'a', 1, 2.5), [1, ])
    self.assertEqual(coreObj.maybeTypes(str, 'a', 1, 'b'), ['a', 'b'])
    # Test with all incompatible values
    self.assertEqual(coreObj.maybeTypes(int, 'a', 'b'), [])
    # Test with no arguments
    self.assertEqual(coreObj.maybeTypes(int), [])

  def test_searchKey(self):
    """Test the searchKey method."""
    coreObj = CoreClass()
    self.assertEqual(coreObj.searchKey('a', 'b', a=1, b=2), 1)
    self.assertIsNone(coreObj.searchKey('c', a=1, b=2))

  def test_searchKeys(self):
    """Test the searchKeys method."""
    coreObj = CoreClass()
    self.assertEqual(coreObj.searchKeys('a', 'b', a=1, b=2), [1, 2])
    self.assertIsNone(coreObj.searchKeys('c', a=1, b=2))

  def test_maybeKey(self):
    """Test the maybeKey method."""
    coreObj = CoreClass()
    self.assertEqual(coreObj.maybeKey('a', 'b', a=1, b=2), 1)
    self.assertEqual(coreObj.maybeKey('c', a=1, b=2, c=3), 3)

  def test_maybeKeys(self):
    """Test the maybeKeys method."""
    coreObj = CoreClass()
    self.assertEqual(coreObj.maybeKeys('a', 'b', a=1, b=2), [1, 2])
    self.assertEqual(coreObj.maybeKeys('c', a=1, b=2, c=3), [3])

  def test_empty(self):
    """Test the empty method."""
    coreObj = CoreClass()
    self.assertTrue(coreObj.empty(None, None))
    self.assertFalse(coreObj.empty(None, 1))

  def test_plenty(self):
    """Test the plenty method."""
    coreObj = CoreClass()
    self.assertTrue(coreObj.plenty(1, 2))
    self.assertFalse(coreObj.plenty(None, 1))



"""WorkToy - MetaClass - AbstractNameSpace
The AbstractNameSpace class provides a class with the minimal
functionality required."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from worktoy.core import Keys, Items, Values, Bases
from worktoy.metaclass import MetaNameSpace


class AbstractNameSpace(dict, metaclass=MetaNameSpace):
  """WorkToy - MetaClass - AbstractNameSpace
  The AbstractNameSpace class provides a class with the minimal
  functionality required."""

  def __init__(self, name: str = None, bases: Bases = None,
               **kwargs) -> None:
    super().__init__()
    self._name = name
    self._bases = bases
    self._kwargs = kwargs
    self._latestErrorGet = None
    self._latestErrorSet = None
    self._log = []
    dict.__setitem__(self, '__name__', '\n%s\n' % self.getModuleName())

  def getModuleName(self) -> str:
    """Getter-function for the module name"""
    return 'WorkToy'

  def __setitem__(self, key: str, val: object, *args, **kwargs) -> None:
    self._log.append(['set', key, val, ])
    dict.__setitem__(self, key, val, )

  def __getitem__(self, key: str, *args, **kwargs) -> object:
    try:
      val = dict.__getitem__(self, key)
      self._recordLog(['get', key, val, ])
      return val
    except KeyError as e:
      self._recordLog(['getError', key, e, ])
      raise e

  def __delitem__(self, key: str) -> None:
    dict.__delitem__(self, key)

  def __contains__(self, key: str) -> bool:
    val = dict.__contains__(self, key)
    return True if val else False

  def keys(self) -> Keys:
    """Implementation of keys"""
    return dict.keys(self)

  def items(self) -> Items:
    """Implementation of keys"""
    return dict.items(self)

  def values(self) -> Values:
    """Implementation of keys"""
    return dict.values(self)

  def getName(self) -> str:
    """Getter-function for name of created class"""
    return self._name

  def getBases(self) -> Bases:
    """Getter-function for name of bases"""
    return self._bases

  def getKwargs(self) -> dict:
    """Getter-function for the keyword arguments"""
    return self._kwargs

  def _recordLog(self, entry: list) -> None:
    """Records the log"""
    self._log.append(entry)

  def getLog(self) -> list:
    """Getter-function for the log"""
    return self._log



"""WorkToy - Descriptors - Attribute
Implementation of descriptor attribute."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Any

from worktoy.worktoyclass import WorkToyClass


class AbstractAttribute(WorkToyClass):
  """WorkToy - Fields - IntAttribute
  Field implementation of integer valued descriptor."""

  @classmethod
  @abstractmethod
  def getClassDefaultValue(cls) -> Any:
    """Getter-function for default value."""

  @abstractmethod
  def getType(self) -> type:
    """Getter-function for value type."""

  def __init__(self, defVal: Any = None, *args, **kwargs) -> None:
    WorkToyClass.__init__(self, *args, **kwargs)
    self._defVal = self.maybe(defVal, self.getClassDefaultValue())
    self._fieldName = None
    self._fieldOwner = None

  def __set_name__(self, cls: type, name: str) -> None:
    """Setter for owner and name."""
    self.setFieldName(name)
    self.setFieldOwner(cls)

  def __get__(self, obj: Any, cls: type) -> Any:
    """Getter."""
    return self.explicitGetter(obj, cls)

  def __set__(self, obj: Any, newValue: Any) -> None:
    """Setter"""
    return self.explicitSetter(obj, newValue)

  def __delete__(self, obj: Any, ) -> None:
    """Setter"""
    return self.explicitDeleter(obj)

  def explicitSetter(self, obj: Any, newValue: Any) -> None:
    """Explicit setter function. """
    return setattr(obj, self.getPrivateName(), newValue)

  def explicitGetter(self, obj: Any, cls: type = None) -> Any:
    """Explicit getter function. """
    return getattr(obj, self.getPrivateName(), )

  def explicitDeleter(self, obj: Any, ) -> None:
    """Explicit deleter function. """
    return delattr(obj, self.getPrivateName(), )

  def getDefaultValue(self) -> Any:
    """Getter-function for instance default value."""
    return self._defVal

  def getPrivateName(self) -> str:
    """Getter-function for the private name"""
    return '_%s' % self.getFieldName()

  def extendInit(self, cls: type) -> type:
    """Extends the '__init__' method on the given class."""

    __original_init__ = getattr(cls, '__init__', None)

    def __new_init__(instance: Any, defVal: Any = None,
                     *args, **kwargs) -> None:
      """New initializer."""
      defVal = self.maybe(defVal, self.getDefaultValue())
      setattr(instance, self.getPrivateName(), defVal)

    if __original_init__ is object.__init__:
      setattr(cls, '__init__', __new_init__)
      return cls

    def __extended_init__(instance: Any, *args, **kwargs) -> None:
      """New initializer."""
      __original_init__(instance, *args, **kwargs)
      __new_init__(instance, *args, **kwargs)

    setattr(cls, '__init__', __extended_init__)
    return cls

  def getFieldName(self) -> str:
    """Getter-function for field name."""
    return self._fieldName

  def setFieldName(self, fieldName: str) -> None:
    """Setter-function for field name."""
    self._fieldName = fieldName

  def getFieldOwner(self) -> type:
    """Getter-function for field owner."""
    return self._fieldOwner

  def setFieldOwner(self, fieldOwner: type) -> None:
    """Setter-function for field owner."""
    self._fieldOwner = self.extendInit(fieldOwner)



"""WorkToy - MetaClass - AbstractMetaClass
This provides the abstract baseclass for metaclasses."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Any

from icecream import ic

from worktoy.core import Bases
from worktoy.metaclass import MetaMetaClass

ic.configureOutput(includeContext=True)


class AbstractMetaClass(MetaMetaClass):
  """WorkToy - MetaClass - AbstractMetaClass
  This provides the abstract baseclass for metaclasses."""

  @classmethod
  @abstractmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> Any:
    """
    The namespace for a newly created class, contains the key-value pairs
    given in the class body. By default, the builtin 'dict' provides the
    datastructure for the namespace. A custom metaclass may provide an
    alternative to this behaviour by reimplementing the __prepare__
    method.

    The default implementation deployed by 'type' and by custom
    metaclasses not implementing __prepare__ is to return an empty
    instance of 'dict'. A custom implementation can achieve enhanced
    functionality by pre-populating the instance of 'dict' or by returning
    an instance of a custom class instead of 'dict'.

    Introducing a custom class in the place of 'dict' provides a powerful
    tool for customizing the class creation process. However, the custom
    class must satisfy the following conditions:

    1.  The object transmitted to the super call in the __new__ method
        must be recognised as belonging to class 'dict' or as a subclass.
    2.  It must implement: __setitem__, __getitem__, __contains__
    3.  IMPORTANT: The __getitem__ implementation must raise a KeyError when
        receiving a key not matching any value. If the custom class does
        not raise a KeyError, HIGHLY UNDEFINED BEHAVIOUR will result.

    Developers implementing a custom class instead of dict for the
    namespace are encouraged to read the above conditions carefully.

    Below is a brief description of the class creation process.
    Parameters
    ----------
    mcls : type
      The metaclass creating the new class.
    cls : type
      The class being created.
    self : object
      Instance created by the newly created class.
    ----------

    The class body is the code creating the class. For example,

    class NewClass(metaclass=MetaClass):
    # Start of class body
      _classVariable = None

      @classmethod
      def getClassVariable(cls) -> object:
        return cls._classVariable

      def __init__(self, *args, **kwargs) -> None:
        self._instanceVariable = math.cos(pi/6)

      def instanceMethod(self) -> object:
        self
    # End of class body

    The code in the body is executed and each line provides an entry to
    the namespace with the left-hand side of the equal side indicating the
    key and the right-hand side the value. This key value pair is then
    placed in the object returned by the __prepare__ method. The namespace
    now populated with the data obtained from the class body becomes
    available to the __new__ method along with the name of the class and a
    tuple containing its baseclasses. Any keyword arguments set in the
    class constructor are also passed along to the __new__ method. (This
    does not include metaclass=MetaClass keyword argument though).

    Once the __new__ method completes, it returns the new class which is
    then passed to the __init__ function on the metaclass. Here, the new
    class is initialised.

    When the class has been initialised, it is also possible to augment the
    instance creation process. When a call is sent to the class,
    it generally creates a new instance of itself. The __call__ method on
    the metaclass defines this procedure, allowing for customization of
    instance creation as well.

    SUMMARY
    __prepare__
      Creates a namespace object by default a dictionary.
    The code in the class body is run line by line, providing entries into
    the namespace object.
    __new__
      Receives the namespace of the class along with name and bases and
      then returns the newly created class.
    __init__
      Receives the newly created class and initialises it.
    __call__
      Receives that class when an attempt is made to instantiate it."""

  #  NO! @classmethod
  def __new__(mcls, name, bases, nameSpace, **kwargs) -> type:
    """DO NOT place class method decorator on the __new__ method. The
    method is automatically a class method, and placing a redundant
    decorator will cause undefined behaviour.
    The baseclass implementation of __new__ collects arguments in the
    MetaClassParams body and passes them to the super call."""
    return super().__new__(mcls, name, bases, nameSpace, **kwargs)

  def __init__(cls, name=None, bases=None, nameSpace=None, **kwargs) -> None:
    """The baseclass implementation uses the MetaClassParams instance to
    collect and pass the arguments to the super call."""
    MetaMetaClass.__init__(cls, name, bases, nameSpace, **kwargs)

  def __call__(cls, *args, **kwargs) -> Any:
    """This method refers to the call method on the class itself, meaning
    instance creation. Below is shown an approximation of the steps
    involved:

      newInstance = newClass(, *args, **kwargs)
      ...
    This line invokes the instance creation in the __call__ method:
      newInstance = NewClass.__new__(NewClass)
      NewClass.__init__(newInstance, *args, **kwargs)
      return newInstance

    Upon completion, the new instance is returned. Please note that this
    is a demonstration. The real implementation may not produce the same
    results."""
    return MetaMetaClass.__call__(cls, *args, **kwargs)



"""WorkToy - Wait A Minute! - TypeSupportError
  This exception should be raised when the argument type is not supported."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.waitaminute import MetaXcept


class TypeSupportError(MetaXcept):
  """WorkToy - Wait A Minute! - TypeSupportError
  This exception should be raised when the argument type is not supported."""

  def __init__(self, expectedType: type, actualValue: Any, argName: str,
               *args, **kwargs) -> None:
    MetaXcept.__init__(self, *args, **kwargs)
    self._expectedType = self.pretty(expectedType)
    self._actualValue = self.pretty(actualValue)
    self._actualType = self.pretty(actualValue.__class__)
    self._argName = self.pretty(argName)

  def __str__(self, ) -> str:
    header = MetaXcept.__str__(self)
    expType = self._expectedType
    actType = self._actualType
    actVal = self._actualValue
    argName = self._argName
    funcName = self.getSourceFunctionName()
    body = """Function '%s' expected the argument: '%s' to be an instance of 
    type: '%s', but received '%s' of type '%s'!"""
    msg = body % (funcName, argName, expType, actVal, actType)
    return '%s\n%s' % (header, self.justify(msg))



"""WorkToy - Core - MetaMetaClass
The meta-metaclass implements __repr__ and __str__ for even the
metaclasses themselves. """
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from icecream import ic

from worktoy.core import Bases

ic.configureOutput(includeContext=True)


class _MetaMetaClass(type):
  """The meta-metaclass implements __repr__ and __str__ for even the
  metaclasses themselves."""

  @classmethod
  def __prepare__(mcls, name, bases, **kwargs) -> dict:
    """Implementing the nameSpace generation"""
    return dict()

  def __new__(mcls,
              name: str,
              bases: Bases,
              nameSpace: dict,
              **kwargs) -> type:
    out = type.__new__(mcls, name, bases, nameSpace, **kwargs)
    if mcls is _MetaMetaClass:
      setattr(out, '__meta_metaclass__', mcls)
    return out

  def __init__(cls,
               name: str,
               bases: Bases,
               nameSpace: dict,
               **kwargs) -> None:
    type.__init__(cls, name, bases, nameSpace, **kwargs)

  def __str__(cls) -> str:
    """String Representation"""
    return cls.__qualname__

  def __repr__(cls) -> str:
    """Code Representation"""
    return cls.__qualname__


class MetaMetaClass(_MetaMetaClass, metaclass=_MetaMetaClass):
  """This class is both derived from and subclass of the _MetaMetaClass."""

  def __new__(mcls,
              name: str = None,
              bases: Bases = None,
              nameSpace: dict = None,
              **kwargs) -> type:
    return _MetaMetaClass.__new__(mcls, name, bases, nameSpace, **kwargs)

  def __init__(cls,
               name: str = None,
               bases: Bases = None,
               nameSpace: dict = None,
               **kwargs) -> None:
    _MetaMetaClass.__init__(cls, name, bases, nameSpace, **kwargs)



"""Test of GuardClass"""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations
import unittest
from worktoy.worktoyclass import GuardClass
from worktoy.core import Function
from worktoy.waitaminute import UnavailableNameException
from worktoy.waitaminute import UnexpectedStateError, TypeSupportError


class TestGuardClass(unittest.TestCase):

  def setUp(self):
    self.guard = GuardClass()

  def test_noneGuard_with_none(self):
    self.assertIsNone(self.guard.noneGuard(None))

  def test_noneGuard_with_non_none(self):
    with self.assertRaises(UnavailableNameException):
      self.guard.noneGuard("Not None", "varName")

  def test_someGuard_with_none(self):
    with self.assertRaises(UnexpectedStateError):
      self.guard.someGuard(None, "varName")

  def test_someGuard_with_non_none(self):
    obj = "Not None"
    self.assertEqual(self.guard.someGuard(obj, "varName"), obj)

  def test_functionGuard_with_function(self):
    def sample_function():
      pass

    self.assertEqual(self.guard.functionGuard(sample_function),
                     sample_function)

  def test_functionGuard_with_non_function(self):
    with self.assertRaises(TypeSupportError):
      self.guard.functionGuard("Not a function", "name")

  def test_intGuard_with_integer(self):
    integer = 42
    self.assertEqual(self.guard.intGuard(integer, "name"), integer)

  def test_intGuard_with_non_integer(self):
    with self.assertRaises(TypeSupportError):
      self.guard.intGuard("Not an integer", "name")

  def test_floatGuard_with_float(self):
    float_value = 3.14
    self.assertEqual(self.guard.floatGuard(float_value, "name"), float_value)

  def test_floatGuard_with_integer(self):
    integer = 42
    self.assertEqual(self.guard.floatGuard(integer, "name"), integer)

  def test_floatGuard_with_non_float_or_integer(self):
    with self.assertRaises(TypeSupportError):
      self.guard.floatGuard("Not a float", "name")

  def test_strGuard_with_string(self):
    text = "Hello, World!"
    self.assertEqual(self.guard.strGuard(text, "name"), text)

  def test_strGuard_with_non_string(self):
    with self.assertRaises(TypeSupportError):
      self.guard.strGuard(42, "name")


if __name__ == '__main__':
  unittest.main()



"""WorkToy - MetaClass
Implementation of metaclass support."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from ._meta_metaclass import MetaMetaClass
from ._abstract_metaclass import AbstractMetaClass
from ._meta_namespace import MetaNameSpace
from ._abstract_namespace import AbstractNameSpace


"""WorkToy - Descriptors - Field
Basic descriptor implementation."""
#  MIT Licence
#  Copyright (c) 2023 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.worktoyclass import WorkToyClass
from worktoy.core import Function


class Field(WorkToyClass):
  """WorkToy - Fields - Field
  Basic descriptor implementation."""

  def __init__(self, *args, **kwargs) -> None:
    WorkToyClass.__init__(self, *args, **kwargs)
    self._defaultValue = (args or (None,))[0]
    self._fieldName = None
    self._fieldOwner = None
    self._getterFunction = None
    self._setterFunction = None
    self._deleterFunction = None

  def __matmul__(self, other: Any) -> Field:
    """Assigns the default value"""
    self._defaultValue = other
    return self

  def _getDefaultValue(self) -> Any:
    """Getter-function for the default value."""
    return self._defaultValue

  def setFieldName(self, fieldName: str) -> None:
    """Setter-function for the field name."""
    self._fieldName = fieldName

  def getFieldName(self, ) -> str:
    """Getter-function for the field name."""
    if self._fieldName is None:
      from worktoy.waitaminute import MissingArgumentException
      raise MissingArgumentException('_fieldName')
    return self._fieldName

  def getPrivateFieldName(self, ) -> str:
    """Getter-function for the private field name on the object. """
    return '_%s' % self.getFieldName()

  def setFieldOwner(self, cls: type) -> None:
    """Setter-function for the field owner."""
    cls = self.wrapFieldOwnerInit(cls)
    self._fieldOwner = cls

  def wrapFieldOwnerInit(self, cls: type) -> type:
    """Wraps the initiator of the field owner."""
    __original_init__ = getattr(cls, '__init__', None)

    if __original_init__ is object.__init__:
      __original_init__ = lambda *args, **kwargs: None

    def __new_init__(instance: object, *args, **kwargs) -> None:
      """Wrapper on the original init."""
      setattr(instance, self.getPrivateFieldName(), self._getDefaultValue())
      if __original_init__ is object.__init__:
        __original_init__(instance)
      else:
        __original_init__(instance, *args, **kwargs)

    setattr(cls, '__init__', __new_init__)
    return cls

  def getFieldOwner(self, ) -> type:
    """Getter-function for the field owner."""
    if self._fieldOwner is None:
      from worktoy.waitaminute import MissingArgumentException
      raise MissingArgumentException('fieldOwner')
    return self._fieldOwner

  def __set_name__(self, cls: type, name: str) -> None:
    """At creation of owner."""
    self.setFieldName(name)
    self.setFieldOwner(cls)

  def __get__(self, obj: object, cls: type) -> Any:
    """Getter descriptor."""
    return self.someGuard(self._getterFunction)(obj)

  def __set__(self, obj: object, newValue: Any) -> None:
    """Setter-descriptor."""
    self.someGuard(self._setterFunction)(obj, newValue)

  def __delete__(self, obj: object) -> None:
    """Deleter-descriptor"""
    self.someGuard(self._deleterFunction)(obj)

  def GET(self, getterFunction: Function) -> Function:
    """Sets the getter function to the decorated function before returning
    it."""
    self.overRideGuard(
      self._getterFunction, '_getterFunction', getterFunction)
    self._getterFunction = getterFunction
    return getterFunction

  def SET(self, setterFunction: Function) -> Function:
    """Sets the setter function to the decorated function before returning
    it."""
    self.overRideGuard(
      self._setterFunction, '_setterFunction', setterFunction)
    self._setterFunction = setterFunction
    return setterFunction

  def DELETER(self, deleterFunction: Function) -> Function:
    """Sets the deleter function to the decorated function before returning
    it."""
    self.overRideGuard(
      self._deleterFunction, '_deleterFunction', deleterFunction)
    self._deleterFunction = deleterFunction
    return deleterFunction
